/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package imcacat.procrun;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.io.ByteArrayInputStream;
import java.util.Arrays;

import org.junit.Test;

/**
 * I test {@link FixedSizeBufferInputStreamConsumer}.
 */
public class FixedSizeBufferInputStreamConsumerTest extends ProcrunTest {
  @Test
  public void testNoInput() throws InterruptedException {
    ByteArrayInputStream stream = new ByteArrayInputStream(new byte[0]);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(128);
    consumer.stream(stream);
    consumer.start();
    consumer.join();
    assertFalse(consumer.hasException());
    byte[] expected = new byte[0];
    byte[] actual = byteBufferToArray(consumer.data());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testWithInput() throws InterruptedException {
    byte[] expected = "abc".getBytes();
    ByteArrayInputStream stream = new ByteArrayInputStream(expected);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(128);
    consumer.stream(stream);
    consumer.start();
    consumer.join();
    assertFalse(consumer.hasException());
    byte[] actual = byteBufferToArray(consumer.data());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputSizeEqualToBufferSize() throws InterruptedException {
    byte[] input = "abc".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer =
        new FixedSizeBufferInputStreamConsumer(input.length);
    consumer.stream(stream);
    consumer.start();
    consumer.join();
    assertFalse(consumer.hasException());
    byte[] expected = input;
    byte[] actual = byteBufferToArray(consumer.data());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer01() throws InterruptedException {
    byte[] input = "abcd".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(3);
    consumer.stream(stream);
    consumer.start();
    consumer.join();
    assertFalse(consumer.hasException());
    byte[] expected = "bcd".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(consumer.data());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer02() throws InterruptedException {
    byte[] input = "abcde".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(3);
    consumer.stream(stream);
    consumer.start();
    consumer.join();
    assertFalse(consumer.hasException());
    byte[] expected = "cde".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(consumer.data());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer03() throws InterruptedException {
    byte[] input = "abcdef".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(3);
    consumer.stream(stream);
    consumer.start();
    consumer.join();
    assertFalse(consumer.hasException());
    byte[] expected = "def".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(consumer.data());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer04() throws InterruptedException {
    byte[] input = "abcdefg".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(3);
    consumer.stream(stream);
    consumer.start();
    consumer.join();
    assertFalse(consumer.hasException());
    byte[] expected = "efg".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(consumer.data());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer05() throws InterruptedException {
    byte[] input = "ab".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(1);
    consumer.stream(stream);
    consumer.start();
    consumer.join();
    assertFalse(consumer.hasException());
    byte[] expected = "b".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(consumer.data());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer06() throws InterruptedException {
    byte[] input = "abc".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(1);
    consumer.stream(stream);
    consumer.start();
    consumer.join();
    assertFalse(consumer.hasException());
    byte[] expected = "c".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(consumer.data());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer07() throws InterruptedException {
    byte[] input = "abcdefghijklmnop".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(6);
    consumer.stream(stream);
    consumer.start();
    consumer.join();
    assertFalse(consumer.hasException());
    byte[] expected = "klmnop".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(consumer.data());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer08() throws InterruptedException {
    byte[] input = "These aren't the droids you're looking for.".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(19);
    consumer.stream(stream);
    consumer.start();
    consumer.join();
    assertFalse(consumer.hasException());
    String expected = "you're looking for.";
    String actual = new String(byteBufferToArray(consumer.data()), US_ASCII);
    assertEquals(expected, actual);
  }

  @Test
  public void testInputBiggerThanBuffer09() throws InterruptedException {
    byte[] input = "These aren't the droids you're looking for.".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(23);
    consumer.stream(stream);
    consumer.start();
    consumer.join();
    assertFalse(consumer.hasException());
    String expected = "ids you're looking for.";
    String actual = new String(byteBufferToArray(consumer.data()), US_ASCII);
    assertEquals(expected, actual);
  }
}
