/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package imcacat.procrun;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.Arrays;

import org.junit.Test;

/**
 * I test {@link ProcessRunner}.
 */
public class ProcessRunnerTest extends ProcrunTest {
  @Test
  public void testExitStatus01() throws InterruptedException {
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("true"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
  }

  @Test
  public void testExitStatus02() throws InterruptedException {
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("false"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertNotEquals(0, result.exitStatus().intValue());
  }

  @Test
  public void testExitStatus03() throws InterruptedException {
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("sh", "-c", "exit 2"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(2, result.exitStatus().intValue());
  }

  @Test
  public void testStdout01() throws InterruptedException {
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("printf", "abc"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    byte[] actual = byteBufferToArray(result.stdout());
    byte[] expected = "abc".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    assertEquals("abc", result.stdoutString());
    assertTrue(result.hasStderr());
    assertEquals(0, result.stderr().remaining());
  }

  @Test
  public void testStdout02() throws InterruptedException {
    ProcessRunnerResult result =
        ProcessRunner.run(new ProcessBuilder("sh", "-c", "sleep 1; exec printf abc"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    byte[] actual = byteBufferToArray(result.stdout());
    byte[] expected = "abc".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    assertEquals("abc", result.stdoutString());
    assertTrue(result.hasStderr());
    assertEquals(0, result.stderr().remaining());
  }

  @Test
  public void testStderr01() throws InterruptedException {
    ProcessRunnerResult result = ProcessRunner.run(
        new ProcessBuilder("sh", "-c", "exec 1>&2; exec printf abc"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    assertEquals(0, result.stdout().remaining());
    byte[] actual = byteBufferToArray(result.stderr());
    byte[] expected = "abc".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    assertEquals("abc", result.stderrString());
  }

  @Test
  public void testStderr02() throws InterruptedException {
    ProcessRunnerResult result = ProcessRunner.run(
        new ProcessBuilder("sh", "-c", "sleep 1; exec 1>&2; exec printf abc"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    assertEquals(0, result.stdout().remaining());
    byte[] actual = byteBufferToArray(result.stderr());
    byte[] expected = "abc".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    assertEquals("abc", result.stderrString());
  }

  @Test
  public void testStdoutAndStderr01() throws InterruptedException {
    ProcessRunnerResult result = ProcessRunner.run(
        new ProcessBuilder("sh", "-c", "printf abc; exec 1>&2; exec printf def"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    byte[] actual;
    byte[] expected;
    actual = byteBufferToArray(result.stdout());
    expected = "abc".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    actual = byteBufferToArray(result.stderr());
    expected = "def".getBytes();
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testStdoutAndStderr02() throws InterruptedException {
    ProcessRunnerResult result = ProcessRunner.run(
        new ProcessBuilder("sh", "-c", "printf abc 1>&2; exec printf def"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    byte[] actual;
    byte[] expected;
    actual = byteBufferToArray(result.stdout());
    expected = "def".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    actual = byteBufferToArray(result.stderr());
    expected = "abc".getBytes();
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testStdoutAndStderr03() throws InterruptedException {
    ProcessRunnerResult result = ProcessRunner.run(
        new ProcessBuilder("sh", "-c", "printf abc; printf def 1>&2; exec printf ghi"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    byte[] actual;
    byte[] expected;
    actual = byteBufferToArray(result.stdout());
    expected = "abcghi".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    actual = byteBufferToArray(result.stderr());
    expected = "def".getBytes();
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testStdoutConsumerException() throws InterruptedException {
    final IOException expected = new IOException("Bad motivator");
    InputStreamConsumer stdoutConsumer = new InputStreamConsumer() {
      @Override
      protected int read() throws IOException {
        throw expected;
      }

      @Override
      public ByteBuffer data() {
        return ByteBuffer.allocate(0);
      }
    };
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("sleep", "5"),
        stdoutConsumer, null);
    assertTrue(result.hasException());
    assertEquals(expected, result.exception());
  }

  @Test
  public void testStderrConsumerException() throws InterruptedException {
    final IOException expected = new IOException("Bad motivator");
    InputStreamConsumer stderrConsumer = new InputStreamConsumer() {
      @Override
      protected int read() throws IOException {
        throw expected;
      }

      @Override
      public ByteBuffer data() {
        return ByteBuffer.allocate(0);
      }
    };
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("sleep", "5"),
        null, stderrConsumer);
    assertTrue(result.hasException());
    assertEquals(expected, result.exception());
  }

  @Test
  public void testStdoutAndStderrConsumerException() throws InterruptedException {
    final IOException stdoutException = new IOException("Bad motivator: stdout");
    final IOException stderrException = new IOException("Bad motivator: stderr");
    InputStreamConsumer stdoutConsumer = new InputStreamConsumer() {
      @Override
      protected int read() throws IOException {
        throw stdoutException;
      }

      @Override
      public ByteBuffer data() {
        return ByteBuffer.allocate(0);
      }
    };
    InputStreamConsumer stderrConsumer = new InputStreamConsumer() {
      @Override
      protected int read() throws IOException {
        throw stderrException;
      }

      @Override
      public ByteBuffer data() {
        return ByteBuffer.allocate(0);
      }
    };
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("sleep", "5"),
        stdoutConsumer, stderrConsumer);
    assertTrue(result.hasException());
    assertTrue(stdoutException == result.exception() || stderrException == result.exception());
  }

  @Test
  public void testInterrupt01() throws InterruptedException {
    ProcessRunner runner = new ProcessRunner(new ProcessBuilder("sleep", "10"));
    runner.start();
    Thread.sleep(1000);
    runner.interrupt();
    long joinStartTime = System.nanoTime();
    runner.join(15000);
    long joinEndTime = System.nanoTime();
    assertTrue(nanosecondsToMilliseconds(joinEndTime - joinStartTime) < 2000);
    ProcessRunnerResult result = runner.result();
    assertTrue(result.hasException());
    assertEquals(InterruptedException.class, result.exception().getClass());
  }

  /*
   * Ensure standard output and standard error consumers die when runner is
   * interrupted.
   */
  @Test
  public void testInterrupt02() throws InterruptedException {
    InputStreamConsumer stdoutConsumer = new FixedSizeBufferInputStreamConsumer(128);
    InputStreamConsumer stderrConsumer = new FixedSizeBufferInputStreamConsumer(128);
    ProcessRunner runner = new ProcessRunner(new ProcessBuilder("sleep", "10"));
    runner.stdoutConsumer(stdoutConsumer);
    runner.stdoutConsumer(stderrConsumer);
    runner.start();
    Thread.sleep(1000);
    runner.interrupt();
    long joinStartTime = System.nanoTime();
    runner.join(3000);
    long joinEndTime = System.nanoTime();
    assertTrue(nanosecondsToMilliseconds(joinEndTime - joinStartTime) < 2000);
    joinStartTime = System.nanoTime();
    stdoutConsumer.join(3000);
    joinEndTime = System.nanoTime();
    assertTrue(nanosecondsToMilliseconds(joinEndTime - joinStartTime) < 2000);
    joinStartTime = System.nanoTime();
    stderrConsumer.join(3000);
    joinEndTime = System.nanoTime();
    assertTrue(nanosecondsToMilliseconds(joinEndTime - joinStartTime) < 2000);
    ProcessRunnerResult result = runner.result();
    assertTrue(result.hasException());
    assertEquals(InterruptedException.class, result.exception().getClass());
  }

  /*
   * Runner is interrupted immediately.
   */
  @Test
  public void testInterrupt03() throws InterruptedException {
    InputStreamConsumer stdoutConsumer = new FixedSizeBufferInputStreamConsumer(128);
    InputStreamConsumer stderrConsumer = new FixedSizeBufferInputStreamConsumer(128);
    ProcessRunner runner = new ProcessRunner(new ProcessBuilder("sleep", "10"));
    runner.stdoutConsumer(stdoutConsumer);
    runner.stdoutConsumer(stderrConsumer);
    runner.start();
    runner.interrupt();
    long joinStartTime = System.nanoTime();
    runner.join(3000);
    long joinEndTime = System.nanoTime();
    assertTrue(nanosecondsToMilliseconds(joinEndTime - joinStartTime) < 2000);
    joinStartTime = System.nanoTime();
    stdoutConsumer.join(3000);
    joinEndTime = System.nanoTime();
    assertTrue(nanosecondsToMilliseconds(joinEndTime - joinStartTime) < 2000);
    joinStartTime = System.nanoTime();
    stderrConsumer.join(3000);
    joinEndTime = System.nanoTime();
    assertTrue(nanosecondsToMilliseconds(joinEndTime - joinStartTime) < 2000);
    ProcessRunnerResult result = runner.result();
    assertTrue(result.hasException());
    assertEquals(InterruptedException.class, result.exception().getClass());
  }

  /*
   * Ensure standard output and standard error consumers die when runner is
   * interrupted, when child process continues to run but has correctly
   * detached from the standard input, output, and error streams.
   */
  @Test
  public void testInterrupt04() throws InterruptedException {
    InputStreamConsumer stdoutConsumer = new FixedSizeBufferInputStreamConsumer(128);
    InputStreamConsumer stderrConsumer = new FixedSizeBufferInputStreamConsumer(128);
    ProcessRunner runner = new ProcessRunner(new ProcessBuilder("sh", "-c",
        "sleep 10 < /dev/null > /dev/null 2>&1"));
    runner.stdoutConsumer(stdoutConsumer);
    runner.stdoutConsumer(stderrConsumer);
    runner.start();
    Thread.sleep(1000);
    runner.interrupt();
    long joinStartTime = System.nanoTime();
    runner.join(3000);
    long joinEndTime = System.nanoTime();
    assertTrue(nanosecondsToMilliseconds(joinEndTime - joinStartTime) < 2000);
    joinStartTime = System.nanoTime();
    stdoutConsumer.join(3000);
    joinEndTime = System.nanoTime();
    assertTrue(nanosecondsToMilliseconds(joinEndTime - joinStartTime) < 2000);
    joinStartTime = System.nanoTime();
    stderrConsumer.join(3000);
    joinEndTime = System.nanoTime();
    assertTrue(nanosecondsToMilliseconds(joinEndTime - joinStartTime) < 2000);
    ProcessRunnerResult result = runner.result();
    assertTrue(result.hasException());
    assertEquals(InterruptedException.class, result.exception().getClass());
  }

  @Test
  public void testRedirectedStderr() throws InterruptedException {
    ProcessBuilder builder = new ProcessBuilder("sh", "-c",
        "printf abc; sleep 1; exec 1>&2; exec printf def");
    builder.redirectErrorStream(true);
    ProcessRunnerResult result = ProcessRunner.run(builder);
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    byte[] actual = byteBufferToArray(result.stdout());
    byte[] expected = "abcdef".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    assertEquals("abcdef", result.stdoutString());
    assertFalse(result.hasStderr());
  }
}
