/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package imcacat.procrun;

import java.io.IOException;

/**
 * I read an {@code InputStream}, placing what I read into a fixed-size
 * buffer, until I encounter the end of the stream or an exception is
 * thrown. If I read more bytes than will fit in the buffer, the buffer will
 * contain the most recent buffer-size bytes read.
 * <p>
 * I check my interrupted status when I can, but I will usually not see that
 * I have been interrupted since I will almost always be blocked reading my
 * {@code InputStream} which is not an interruptible channel. So, the best
 * way to get me to die is to close the {@code InputStream} I am reading.
 */
public class FixedSizeBufferInputStreamConsumer extends BufferInputStreamConsumer {
  private boolean bufferWrapped;

  /**
   * Creates an instance with the specified buffer size. Calls
   * {@code super(bufferSize)}.
   *
   * @param bufferSize size of the buffer into which I read data from my
   *          input stream
   */
  public FixedSizeBufferInputStreamConsumer(int bufferSize) {
    super(bufferSize);
  }

  /**
   * Creates an instance with the specified buffer size and thread name.
   * Calls {@code super(bufferSize, name)}.
   *
   * @param bufferSize size of the buffer into which I read data from my
   *          input stream
   * @param name thread name
   */
  public FixedSizeBufferInputStreamConsumer(int bufferSize, String name) {
    super(bufferSize, name);
  }

  /**
   * Creates an instance with the specified buffer size and thread group.
   * Calls {@code super(bufferSize, group)}.
   *
   * @param bufferSize size of the buffer into which I read data from my
   *          input stream
   * @param group thread group
   */
  public FixedSizeBufferInputStreamConsumer(int bufferSize, ThreadGroup group) {
    super(bufferSize, group);
  }

  /**
   * Creates an instance with the specified buffer size, thread group, and
   * thread name. Calls {@code super(bufferSize, group, name)}.
   *
   * @param bufferSize size of the buffer into which I read data from my
   *          input stream
   * @param group thread group
   * @param name thread name
   */
  public FixedSizeBufferInputStreamConsumer(int bufferSize, ThreadGroup group, String name) {
    super(bufferSize, group, name);
  }

  /**
   * Creates an instance with the specified buffer size, thread group,
   * thread name, and thread stack size. Calls
   * {@code super(bufferSize, group, name, stackSize)}.
   *
   * @param bufferSize size of the buffer into which I read data from my
   *          input stream
   * @param group thread group
   * @param name thread name
   * @param stackSize thread stack size
   */
  public FixedSizeBufferInputStreamConsumer(int bufferSize, ThreadGroup group, String name,
      long stackSize) {
    super(bufferSize, group, name, stackSize);
  }

  @Override
  protected void initialize(int bufferSize) {
    super.initialize(bufferSize);
    bufferWrapped = false;
  }

  @Override
  protected int read() throws IOException {
    int numBytesRead = stream.read(buffer, bufferPosition, buffer.length - bufferPosition);
    if (numBytesRead != -1) {
      bufferPosition += numBytesRead;
      if (bufferPosition == buffer.length) {
        bufferWrapped = true;
        bufferPosition = 0;
      }
    }
    return numBytesRead;
  }

  @Override
  protected void postProcessData() {
    if (!bufferWrapped) return;
    if (bufferPosition == buffer.length) return;
    if (bufferPosition == 0) {
      bufferPosition = buffer.length;
      return;
    }
    if (bufferPosition < (buffer.length / 2)) {
      byte[] tmp = new byte[bufferPosition];
      System.arraycopy(buffer, 0, tmp, 0, tmp.length);
      System.arraycopy(buffer, bufferPosition, buffer, 0, buffer.length - bufferPosition);
      System.arraycopy(tmp, 0, buffer, buffer.length - bufferPosition, tmp.length);
    } else {
      byte[] tmp = new byte[buffer.length - bufferPosition];
      System.arraycopy(buffer, bufferPosition, tmp, 0, tmp.length);
      System.arraycopy(buffer, 0, buffer, tmp.length, bufferPosition);
      System.arraycopy(tmp, 0, buffer, 0, tmp.length);
    }
    bufferPosition = buffer.length;
  }
}
