/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package imcacat.procrun;

import java.io.IOException;
import java.util.Arrays;

/**
 * I read an {@code InputStream}, placing what I read into an automatically
 * growing buffer, until I encounter the end of the stream or an exception
 * is thrown.
 * <p>
 * I check my interrupted status when I can, but I will usually not see that
 * I have been interrupted since I will almost always be blocked reading my
 * {@code InputStream} which is not an interruptible channel. So, the best
 * way to get me to die is to close the {@code InputStream} I am reading.
 */
public class GrowingBufferInputStreamConsumer extends BufferInputStreamConsumer {
  /**
   * Default initial buffer size in bytes. Its value is
   * {@value #DEFAULT_BUFFER_SIZE}.
   */
  public static final int DEFAULT_BUFFER_SIZE = 32;

  /**
   * Creates an instance with an initial buffer size of
   * {@value #DEFAULT_BUFFER_SIZE} bytes. Behaves just like calling the
   * {@code GrowingBufferInputStreamConsumer(int)} constructor with an
   * argument of <code>{@value #DEFAULT_BUFFER_SIZE}</code>.
   */
  public GrowingBufferInputStreamConsumer() {
    this(DEFAULT_BUFFER_SIZE);
  }

  /**
   * Creates an instance with the specified buffer size. Calls
   * {@code super(bufferSize)}.
   *
   * @param bufferSize initial size of the buffer into which I read data
   *          from my input stream
   */
  public GrowingBufferInputStreamConsumer(int bufferSize) {
    super(bufferSize);
  }

  /**
   * Creates an instance with the specified buffer size and thread name.
   * Calls {@code super(bufferSize, name)}.
   *
   * @param bufferSize initial size of the buffer into which I read data
   *          from my input stream
   * @param name thread name
   */
  public GrowingBufferInputStreamConsumer(int bufferSize, String name) {
    super(bufferSize, name);
  }

  /**
   * Creates an instance with the specified buffer size and thread group.
   * Calls {@code super(bufferSize, group)}.
   *
   * @param bufferSize initial size of the buffer into which I read data
   *          from my input stream
   * @param group thread group
   */
  public GrowingBufferInputStreamConsumer(int bufferSize, ThreadGroup group) {
    super(bufferSize, group);
  }

  /**
   * Creates an instance with the specified buffer size, thread group, and
   * thread name. Calls {@code super(bufferSize, group, name)}.
   *
   * @param bufferSize initial size of the buffer into which I read data
   *          from my input stream
   * @param group thread group
   * @param name thread name
   */
  public GrowingBufferInputStreamConsumer(int bufferSize, ThreadGroup group, String name) {
    super(bufferSize, group, name);
  }

  /**
   * Creates an instance with the specified buffer size, thread group,
   * thread name, and thread stack size. Calls
   * {@code super(bufferSize, group, name, stackSize)}.
   *
   * @param bufferSize initial size of the buffer into which I read data
   *          from my input stream
   * @param group thread group
   * @param name thread name
   * @param stackSize thread stack size
   */
  public GrowingBufferInputStreamConsumer(int bufferSize, ThreadGroup group, String name,
      long stackSize) {
    super(bufferSize, group, name, stackSize);
  }

  @Override
  protected int read() throws IOException {
    int numBytesRead = stream.read(buffer, bufferPosition, buffer.length - bufferPosition);
    if (numBytesRead != -1) {
      bufferPosition += numBytesRead;
      if (bufferPosition == buffer.length) {
        buffer = Arrays.copyOf(buffer, buffer.length * 2);
      }
    }
    return numBytesRead;
  }
}
