/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package imcacat.procrun;

import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.util.concurrent.BlockingQueue;

/**
 * I am the base class for all input stream consumers; I read an
 * {@code InputStream} until encountering the end of the stream or an
 * exception is thrown. If my finished-channel is not {@code null}, I offer
 * myself on it before I die.
 * <p>
 * I check my interrupted status when I can, but I will usually not see that
 * I have been interrupted since I will almost always be blocked reading my
 * {@code InputStream} which is not an interruptible channel. So, the best
 * way to get me to die is to close the {@code InputStream} I am reading.
 */
public abstract class InputStreamConsumer extends Thread {
  protected InputStream stream;
  protected BlockingQueue<InputStreamConsumer> finishedChannel;
  protected volatile Throwable exception;

  /**
   * Creates an instance. Calls {@code super()}.
   */
  public InputStreamConsumer() {
    super();
    initialize();
  }

  /**
   * Creates an instance with the specified thread name. Calls
   * {@code super(name)}.
   *
   * @param name thread name
   */
  public InputStreamConsumer(String name) {
    super(name);
    initialize();
  }

  /**
   * Creates an instance with the specified thread group. Calls
   * {@code super(group, (Runnable)null)}.
   *
   * @param group thread group
   */
  public InputStreamConsumer(ThreadGroup group) {
    super(group, (Runnable)null);
    initialize();
  }

  /**
   * Creates an instance with the specified thread group and thread name.
   * Calls {@code super(group, name)}.
   *
   * @param group thread group
   * @param name thread name
   */
  public InputStreamConsumer(ThreadGroup group, String name) {
    super(group, name);
    initialize();
  }

  /**
   * Creates an instance with the specified thread group, thread name, and
   * thread stack size. Calls {@code super(group, null, name, stackSize)}.
   *
   * @param group thread group
   * @param name thread name
   * @param stackSize thread stack size
   */
  public InputStreamConsumer(ThreadGroup group, String name, long stackSize) {
    super(group, null, name, stackSize);
    initialize();
  }

  /**
   * Initializes me. Called from my constructors.
   */
  protected void initialize() {
    stream = null;
    finishedChannel = null;
    exception = null;
  }

  @Override
  public void run() {
    try {
      for (;;) {
        if (isInterrupted()) throw new InterruptedException();
        if (read() != -1) continue;
        postProcessData();
        break;
      }
    } catch (Throwable e) {
      exception = e;
    } finally {
      if (finishedChannel != null) {
        finishedChannel.offer(this);
      }
    }
  }

  /**
   * Returns the finished-channel I use to signal that I have finished
   * running.
   * <p>
   * It is only safe to invoke this method before I have been started.
   *
   * @return finished-channel or {@code null} if one has not been set
   */
  public BlockingQueue<InputStreamConsumer> finishedChannel() {
    return finishedChannel;
  }

  /**
   * Sets the finished-channel I use to signal that I have finished running.
   * <p>
   * It is only safe to invoke this method before I have been started.
   *
   * @param newFinishedChannel new finished-channel or {@code null} if none
   */
  public void finishedChannel(BlockingQueue<InputStreamConsumer> newFinishedChannel) {
    finishedChannel = newFinishedChannel;
  }

  /**
   * Returns the input stream I read from.
   * <p>
   * It is only safe to invoke this method before I have been started.
   *
   * @return input stream or {@code null} if one has not been set
   */
  public InputStream stream() {
    return stream;
  }

  /**
   * Sets the input stream I read from.
   * <p>
   * It is only safe to invoke this method before I have been started.
   *
   * @param newStream new input stream or {@code null} if none
   */
  public void stream(InputStream newStream) {
    stream = newStream;
  }

  /**
   * Tests whether I have an exception. I have an exception if an exception
   * occurred while I was running.
   *
   * @return {@code true} if I have an exception; {@code false} otherwise
   */
  public boolean hasException() {
    return exception != null;
  }

  /**
   * Returns my exception.
   *
   * @return exception or {@code null} if none
   */
  public Throwable exception() {
    return exception;
  }

  /**
   * Returns the data consumed from my input stream.
   * <p>
   * This method can be called safely from another thread only after
   * ensuring all my actions <i>happen-before</i> the action of calling this
   * method from the other thread. This is the Java memory model
   * <i>happens-before</i> relationship. For example, after I have been
   * started, it would be safe to call this method from another thread after
   * it called my {@code join} method.
   *
   * @return consumed data
   */
  public abstract ByteBuffer data();

  /**
   * Reads a chunk of data from my input stream. May block.
   *
   * @return number of bytes read or -1 indicating end-of-stream
   *
   * @throws IOException if an I/O error occurs
   */
  protected abstract int read() throws IOException;

  /**
   * Performs post-processing of data, and is invoked from my {@code run()}
   * method after a call to {@link #read()} returns indicating
   * end-of-stream.
   *
   * This method does nothing, but subclasses may override it to do
   * something.
   */
  protected void postProcessData() {
  }
}