/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package imcacat.procrun;

import java.nio.ByteBuffer;
import java.nio.charset.Charset;

/**
 * I am the result of running a {@link ProcessRunner}.
 */
public class ProcessRunnerResult {
  private final Integer exitStatus;
  private final Throwable exception;
  private final ByteBuffer stdout;
  private final ByteBuffer stderr;

  /**
   * Creates an instance.
   *
   * @param exitStatus exit status of the process or {@code null} if none
   * @param exception exception causing abnormal termination of the
   *          {@code ProcessRunner} or {@code null} if none
   * @param stdout data read from the standard-output stream of the process
   *          or {@code null} if none
   * @param stderr data read from the standard-error stream of the process
   *          or {@code null} if none
   */
  public ProcessRunnerResult(Integer exitStatus, Throwable exception, ByteBuffer stdout,
      ByteBuffer stderr) {
    super();
    this.exitStatus = exitStatus;
    this.exception = exception;
    this.stdout = stdout;
    this.stderr = stderr;
  }

  /**
   * Tests whether I have an exit status.
   *
   * @return {@code true} if I have an exit status; {@code false} otherwise
   */
  public boolean hasExitStatus() {
    return exitStatus != null;
  }

  /**
   * Returns my exit status.
   *
   * @return exit status or {@code null} if none
   */
  public Integer exitStatus() {
    return exitStatus;
  }

  /**
   * Tests whether I have an exception.
   *
   * @return {@code true} if I have an exception; {@code false} otherwise
   */
  public boolean hasException() {
    return exception != null;
  }

  /**
   * Returns my exception.
   *
   * @return exception or {@code null} if none
   */
  public Throwable exception() {
    return exception;
  }

  /**
   * Tests whether I have standard-output data.
   *
   * @return {@code true} if I have standard-output data; {@code false}
   *         otherwise
   */
  public boolean hasStdout() {
    return stdout != null;
  }

  /**
   * Returns my standard-output data.
   * <p>
   * The data is returned as a {@code ByteBuffer} with its limit set to the
   * size of the data. The {@code ByteBuffer} is backed by an array, so
   * calling {@link ByteBuffer#array()} will always succeed.
   *
   * @return standard-output data or {@code null} if none
   */
  public ByteBuffer stdout() {
    return stdout;
  }

  /**
   * Returns my standard-output data as a string decoded using the default
   * character set.
   *
   * @return standard-output string or {@code null} if none
   */
  public String stdoutString() {
    return stdoutString(Charset.defaultCharset());
  }

  /**
   * Returns my standard-output data as a string decoded using character set
   * {@code cs}.
   *
   * @param cs character set to decode data with
   *
   * @return standard-output string or {@code null} if none
   */
  public String stdoutString(Charset cs) {
    if (stdout == null) return null;
    return new String(stdout.array(), 0, stdout.limit(), cs);
  }

  /**
   * Tests whether I have standard-error data.
   *
   * @return {@code true} if I have standard-error data; {@code false}
   *         otherwise
   */
  public boolean hasStderr() {
    return stderr != null;
  }

  /**
   * Returns my standard-error data.
   * <p>
   * The data is returned as a {@code ByteBuffer} with its limit set to the
   * size of the data. The {@code ByteBuffer} is backed by an array, so
   * calling {@link ByteBuffer#array()} will always succeed.
   *
   * @return standard-error data or {@code null} if none
   */
  public ByteBuffer stderr() {
    return stderr;
  }

  /**
   * Returns my standard-error data as a string decoded using the default
   * character set.
   *
   * @return standard-error string or {@code null} if none
   */
  public String stderrString() {
    return stderrString(Charset.defaultCharset());
  }

  /**
   * Returns my standard-error data as a string decoded using character set
   * {@code cs}.
   *
   * @param cs character set to decode data with
   *
   * @return standard-error string or {@code null} if none
   */
  public String stderrString(Charset cs) {
    if (stderr == null) return null;
    return new String(stderr.array(), 0, stderr.limit(), cs);
  }
}
